"""
See "Optimal Carpet Cutting" by Andreas Schutt, Peter J. Stuckey, and Andrew R. Verden, CP'11
See Minizinc challenges 2011, 2012, 2016 and 2021 
"""

from pycsp3 import *

rollWidth, maxRollLength, roomCarpets, rectangles, stairCarpets = data
roomCarpetRectangles, possibleRotations, maximalLengths, maximalWidths = zip(*roomCarpets)
rectangleLengths, rectangleWidths, xOffsets, yOffsets = zip(*rectangles)
stairCarpetLengths, stairCarpetWidths, nCoveredSteps, minCutSteps, maxCuts = zip(*stairCarpets) if len(stairCarpets) > 0 else ([], [], [], [], [])
nRoomCarpets, nRectangles, nStairCarpets = len(roomCarpets), len(rectangles), len(stairCarpets)

Rotations = range(4)  # 0 - 0, 1 - 90, 2 - 180, and 3 - 270

stairCarpetOffsets = [sum(nCoveredSteps[:i]) for i in range(nStairCarpets)]
stairCarpetRanges = [range(stairCarpetOffsets[i], stairCarpetOffsets[i] + nCoveredSteps[i]) for i in range(nStairCarpets)]

nSteps = sum(nCoveredSteps)
stepLengths = [stairCarpetLengths[i] // nCoveredSteps[i] for i in range(nStairCarpets) for _ in range(nCoveredSteps[i])]
stepWidths = [stairCarpetWidths[i] for i in range(nStairCarpets) for _ in range(nCoveredSteps[i])]

totalArea = sum(l * w for (l, w, _, _) in rectangles) + sum(stairCarpetLengths[i] * stairCarpetWidths[i] for i in range(nStairCarpets))
minRollLength = (totalArea // rollWidth) + (1 if totalArea % rollWidth > 0 else 0)
totalLength = sum(max(maximalLengths[i], maximalWidths[i]) for i in range(nRoomCarpets)) + sum(stairCarpetLengths)
maxRollLength = min(maxRollLength, totalLength)

rangeRectangleSizes = range(min(min(l, w) for (l, w, _, _) in rectangles), max(max(l, w) for (l, w, _, _) in rectangles) + 1)

nexts = [next(j for j in range(nRoomCarpets) if i in roomCarpetRectangles[j]) for i in range(nRectangles)]

# z is the carpet roll length
z = Var(range(minRollLength, maxRollLength + 1))

# x[i] is the x-coordinate of the ith room carpet
x = VarArray(size=nRoomCarpets, dom=range(maxRollLength + 1))

# y[i] is the y-coordinate of the ith room carpet
y = VarArray(size=nRoomCarpets, dom=range(rollWidth + 1))

# r[i] is the rotation of the ith room carpet
r = VarArray(size=nRoomCarpets, dom=Rotations)

# r0or180[is] is 1 if the the rotation of the ith room carpet is 0 or 180 degrees
r0or180 = VarArray(size=nRoomCarpets, dom={0, 1})

# r0or90[is] is 1 if the the rotation of the ith room carpet is 0 or 90 degrees
r0or90 = VarArray(size=nRoomCarpets, dom={0, 1})

# xr[j] is the x-coordinate of the jth rectangle
xr = VarArray(size=nRectangles, dom=range(maxRollLength + 1))

# yr[j] is the y-coordinate of the jth rectangle
yr = VarArray(size=nRectangles, dom=range(rollWidth + 1))

# lr[j] is the length of the jth rectangle (considering a possible rotation)
lr = VarArray(size=nRectangles, dom=rangeRectangleSizes)

# wr[j] is the width of the jth rectangle (considering a possible rotation)
wr = VarArray(size=nRectangles, dom=rangeRectangleSizes)

if nSteps > 0:
    # xs[k] is the x-coordinate of the kth step of the stair
    xs = VarArray(size=nSteps, dom=range(maxRollLength + 1))

    # ys[k] is the y-coordinate of the kth step of the stair
    ys = VarArray(size=nSteps, dom=range(rollWidth + 1))

    # lp[i][j] is 1 if the jth covered step by the ith stair carpet is the last step of a part (of the partition of the stair carpet)
    lp = VarArray(size=[nStairCarpets, max(nCoveredSteps)], dom=lambda i, j: {0, 1} if j < nCoveredSteps[i] else None)

    ls, ws = stepLengths, stepWidths

else:
    xs = ys = lp = ls = ws = []

X, Y, L, W = xr + xs, yr + ys, lr + ls, wr + ws

satisfy(
    # computing the lengths of rectangles
    [lr[i] == rectangleWidths[i] + (rectangleLengths[i] - rectangleWidths[i]) * r0or180[nexts[i]] for i in range(nRectangles)],

    # computing the widths of rectangles
    [wr[i] == rectangleLengths[i] + (rectangleWidths[i] - rectangleLengths[i]) * r0or180[nexts[i]] for i in range(nRectangles)],

    # enforcing room carpets to stay within limits
    [
        [x[i] + maximalWidths[i] + (maximalLengths[i] - maximalWidths[i]) * r0or180[i] <= z for i in range(nRoomCarpets)],
        [y[i] + maximalLengths[i] + (maximalWidths[i] - maximalLengths[i]) * r0or180[i] <= rollWidth for i in range(nRoomCarpets)]
    ],

    # enforcing rectangles (of room carpets) to stay within limits
    [
        [xr[j] + lr[j] <= z for i in range(nRoomCarpets) for j in roomCarpetRectangles[i]],
        [yr[j] + wr[j] <= rollWidth for i in range(nRoomCarpets) for j in roomCarpetRectangles[i]]
    ],

    # respecting possible rotations of room carpets
    [r[i] in possibleRotations[i] for i in range(nRoomCarpets)],

    # determining when rotations are 0 or 90 degrees
    [r0or90[i] == (r[i] in {0, 1}) for i in range(nRoomCarpets)],

    # determining when rotations are 0 or 180 degrees
    [r0or180[i] == (r[i] in {0, 2}) for i in range(nRoomCarpets)],

    # computing the x-coordinates of the rectangles in room carpets
    [xr[j] == x[i] + xOffsets[j][r[i]] for i in range(nRoomCarpets) for j in roomCarpetRectangles[i]],

    # computing the y-coordinates of the rectangles in room carpets
    [yr[j] == y[i] + yOffsets[j][r[i]] for i in range(nRoomCarpets) for j in roomCarpetRectangles[i]],

    # enforcing stair steps to stay within limits
    [
        [xs[j] + stepLengths[j] <= z for i in range(nStairCarpets) for j in stairCarpetRanges[i]],
        [ys[j] + stepWidths[j] <= rollWidth for i in range(nStairCarpets) for j in stairCarpetRanges[i]]
    ],

    # breaking symmetries between steps of a stair carpet  tag(symmetry-breaking)
    [
        [ys[j] <= ys[j + 1] for i in range(nStairCarpets) for j in stairCarpetRanges[i]
         if j + 1 in stairCarpetRanges[i]],
        [(ys[j] < ys[j + 1]) | (xs[j] + stepLengths[j] <= xs[j + 1]) for i in range(nStairCarpets) for j in stairCarpetRanges[i]
         if j + 1 in stairCarpetRanges[i]]
    ],

    # computing the last steps in the parts of the partitions of the stair carpets
    [
        [lp[i][nCoveredSteps[i] - 1] == 1 for i in range(nStairCarpets)],

        [lp[i][j] == (ys[j + offset] < ys[j + offset + 1]) | (xs[j + offset] + stepLengths[j] < xs[j + offset + 1])
         for i, offset in enumerate(stairCarpetOffsets) for j in range(nCoveredSteps[i] - 1)],

        [(xs[j + offset] + 2 * stepLengths[j + offset] <= z) | (lp[i][j] == 1) for i, offset in enumerate(stairCarpetOffsets) for j in
         range(nCoveredSteps[i] - 1)],

        [Sum(lp[i][:nCoveredSteps[i]]) <= maxCuts[i] + 1 for i in range(nStairCarpets)],

        [lp[i][j] == 0 for i in range(nStairCarpets) if minCutSteps[i] > 1 for j in range(minCutSteps[i] - 1)],

        [(lp[i][j] == 0) | (lp[i][k] == 0) for i in range(nStairCarpets) if minCutSteps[i] > 1 for j in range(minCutSteps[i] - 1, nCoveredSteps[i])
         for k in range(j - minCutSteps[i] + 1, j)]
    ],

    # respecting roll length
    Cumulative(origins=X, lengths=L, heights=W) <= rollWidth,

    # respecting roll width
    Cumulative(origins=Y, lengths=W, heights=L) <= z,

    # non-overlapping
    NoOverlap(origins=zip(X, Y), lengths=zip(L, W))
)

minimize(
    # minimizing the carpet roll length
    z
)
