"""
See https://opthub.uniud.it/problem/timetabling/gbac
See Luca Di Gaspero, Andrea Schaerf: "Hybrid Local Search Techniques for the Generalized Balanced Academic Curriculum Problem". Hybrid Metaheuristics 2008: 146-157
See https://www.csplib.org/Problems/prob064/

Examples of Execution:
  python GBACP.py -data=UD01.json
  python GBACP.py -variant=table -data=UD01.json
"""

from pycsp3 import *

nYears, nPeriodsPerYear, loadBounds, courseLoads, curricula, precedences, undesiredPeriods = data
nPeriods, nCourses, nCurricula = nYears * nPeriodsPerYear, len(courseLoads), len(curricula)

max_load = sum(courseLoads)
total_load = [sum(courseLoads[i] for i in c) for c in curricula]
ideal_floor = [total_load[c] // nPeriods for c in range(nCurricula)]
ideal_ceil = [total_load[c] // nPeriods + (0 if total_load[c] % nPeriods == 0 else 1) for c in range(nCurricula)]
distinctCurricula = [curricula[i] for i in range(nCurricula) if all(curricula[j] != curricula[i] for j in range(i))]

# x[i] is the period for the ith course
x = VarArray(size=nCourses, dom=range(nPeriods))

# y[c][p] is the load in period p of curriculum c
y = VarArray(size=[nCurricula, nPeriods], dom=range(max_load + 1))

# d[c][p] is the difference (delta) between the ideal load and the effective load in period p of curriculum c
d = VarArray(size=[nCurricula, nPeriods], dom=range(max_load + 1))

satisfy(
    # respecting authorized loads of courses for all periods and curricula
    [Cardinality([x[i] for i in curriculum], occurrences={p: range(loadBounds.min, loadBounds.max + 1) for p in range(nPeriods)})
     for curriculum in distinctCurricula],

    # respecting prerequisites
    [x[i] < x[j] for i, j in precedences],

    # computing loads
    [BinPacking([x[i] for i in curricula[c]], sizes=[courseLoads[i] for i in curricula[c]], loads=y[c]) for c in range(nCurricula)]
)

if not variant():
    satisfy(
        # computing deltas
        d[c][p] == Maximum(y[c][p] - ideal_ceil[c], ideal_floor[c] - y[c][p]) for c in range(nCurricula) for p in range(nPeriods)
    )

elif variant("table"):
    T = [[(max(v - ideal_ceil[c], ideal_floor[c] - v), v) for v in range(max_load + 1)] for c in range(nCurricula)]

    satisfy(
        # computing deltas
        (d[c][p], y[c][p]) in T[c] for c in range(nCurricula) for p in range(nPeriods)
    )

minimize(
    # minimizing preference violations and unbalanced loads
    Sum(d[c][p] * d[c][p] for c in range(nCurricula) for p in range(nPeriods))
    + Sum(x[i] == v + k * nPeriodsPerYear for (i, v) in undesiredPeriods for k in range(nPeriodsPerYear))
)


