"""
See M. Wallace and N. Yorke-Smith. A new constraint programming model and solving for the cyclic hoist scheduling problem. Constraints, 25(3-4):319–337, 2020
    A. El Amraoui and M. Elhafsi. An efficient new heuristic for the hoist scheduling problem. Computers & Operations Research, 67:184–192, 2016.

Examples of Execution:
  python HSP.py -data=10405.json
  python HSP.py -variant=aux -data=10405.json
  python HSP.py -variant=table -data=10405.json
"""
  
from pycsp3 import *

intervals, ld, ud = data  # loaded and unloaded time for one move (from one tank to one of its successors)

min_dips, max_dips = zip(*intervals)  # min and max dip durations
nTanks = len(intervals) + 1
horizon = sum(min_dips) + ld * nTanks + ud * nTanks + 1

# t[i] is the time when the product is picked from the ith tank
t = VarArray(size=nTanks, dom=range(horizon - ld - ud))

# o[i] is the order in which the picking operation of the ith tank is executed
o = VarArray(size=nTanks, dom=range(nTanks))

# d[i] is the dip duration in the ith tank (plus ld)
d = VarArray(size=nTanks, dom=lambda i: range(min_dips[i - 1] + ld, max_dips[i - 1] + 1 + ld) if i != 0 else {0})  # i-1 because of special tank 0

# time of the cycle
z = Var(range(nTanks * (ld + ud), horizon))


def duration_limit(i, j, decreasing):
    assert i < j
    return ld + ud * (abs(j - i) + (1 if decreasing else - 1))


satisfy(
    # the first operation corresponds to starting the cycle by putting a product in the first tank
    [t[0] == 0, o[0] == 0],

    # taking into account the time of going back to tank 0
    [t[i] <= horizon - ld - (i + 1) * ud for i in range(1, nTanks)],

    # operations are executed in some order
    AllDifferent(o),

    # order of picking operations must be compatible with their picking times
    [(o[i] < o[j]) == (t[i] < t[j]) for i, j in combinations(nTanks, 2)],

    # ensuring a minimal duration between any two tanks
    NoOverlap(origins=t, lengths=ld + min(ud, min(min_dips))),

    # computing the time of the cycle
    z == Maximum(t[i] + ld + (i + 1) * ud for i in range(1, nTanks)),

    # ensuring a duration limit between any two tanks
    [
        [imply(o[j] - o[i] == 1, t[j] - t[i] >= duration_limit(i, j, False)) for i, j in combinations(nTanks, 2)],
        [imply(o[i] - o[j] == 1, t[i] - t[j] >= duration_limit(i, j, True)) for i, j in combinations(nTanks, 2)]
    ]
)

if not variant():
    satisfy(
        # computing the dip duration of each product
        d[i] == ift(o[i] < o[i - 1], z + t[i] - t[i - 1], t[i] - t[i - 1]) for i in range(1, nTanks)
    )
else:
    assert variant() in ("aux", "table")

    # g[i] is the gap (distance) between t[i] and t[i+1]
    g = VarArray(size=nTanks - 1, dom=range(-horizon, horizon))

    satisfy(
        # reducing the domain of auxiliary variables
        [
            [g[i] < horizon - ld - (i + 2) * ud for i in range(1, nTanks - 1)],
            [g[i] > -horizon + ld + (i + 2) * ud for i in range(1, nTanks - 1)]
        ]
    )

    if variant("aux"):

        satisfy(
            # computing the gap/distance between the picking time concerning two successive tanks
            [g[i] == t[i + 1] - t[i] for i in range(nTanks - 1)],

            # computing the dip duration of each product
            [d[i] == ift(g[i - 1] < 0, z + g[i - 1], g[i - 1]) for i in range(1, nTanks)]
        )

    elif variant("table"):

        T1 = [(v2 - v1, v2, v1) for v2 in t[2].dom for v1 in t[1].dom if v2 - v1 in g[1].dom]
        T2 = [[(vt + vg if vg < 0 else vg, vg, vt) for vg in g[i - 1].dom for vt in z.dom if (vt + vg if vg < 0 else vg) in d[i].dom] for i in range(1, nTanks)]

        satisfy(
            # computing the gap/distance between two successive tanks
            [(g[i], t[i + 1], t[i]) in T1 for i in range(nTanks - 1)],

            # computing the dip duration of each product
            [(d[i], g[i - 1], z) in T2[i - 1] for i in range(1, nTanks)]
        )

minimize(
    #  minimizing the time of the cycle
    z
)

