"""
See Marzio De Biasi: "Binary Puzzle is NP-complete"

Example of Execution:
  python BinaryPuzzle.py -data=60
  python BinaryPuzzle.py -variant=regular -data=60
"""

from pycsp3 import *

n = data
assert n % 2 == 0
m = n // 2

# x[i][j] is the value in the cell of the grid at coordinates (i,j)
x = VarArray(size=[n, n], dom={0, 1})

if not variant():
    satisfy(
        # ensuring the same number of 0s and 1s in rows
        [Sum(x[i]) == m for i in range(n)],

        # ensuring the same number of 0s and 1s in columns
        [Sum(x[:, j]) == m for j in range(n)],

        # forbidding sequences of 3 consecutive 0s or 1s in rows
        [Sum(x[i, j:j + 3]) in range(1, 3) for i in range(n) for j in range(n - 2)],

        # forbidding sequences of 3 consecutive 0s or 1s in columns
        [Sum(x[i:i + 3, j]) in range(1, 3) for j in range(n) for i in range(n - 2)]
    )

elif variant("regular"):
    pairs = [(j, k) for j in range(3) for k in range(3) if (j == 0 and k > 0) or (j > 0 and k == 0)]

    q = Automaton.q
    t = [(q(0, 0, 0), 0, q(0, 1, 0)), (q(0, 0, 0), 1, (q(1, 0, 1)))] \
        + [(q(i, j, k), 0, q(i, j + 1, 0)) for i in range(m + 1) for j, k in pairs if j < 2] \
        + [(q(i, j, k), 1, q(i + 1, 0, k + 1)) for i in range(m) for j, k in pairs if k < 2]
    A = Automaton(start=q(0, 0, 0), final=[q(m, j, k) for j, k in pairs], transitions=t)

    satisfy(
        # ensuring the same number of 0s and 1s in rows, while forbidding sequences of 3 consecutive 0s or 1s
        [x[i] in A for i in range(n)],

        # ensuring the same number of 0s and 1s in columns, while forbidding sequences of 3 consecutive 0s or 1s
        [x[:, j] in A for j in range(n)]
    )

satisfy(
    # forbidding identical rows
    AllDifferentList(x[i] for i in range(n)),  #.to_table(),

    # forbidding identical columns
    AllDifferentList(x[:, j] for j in range(n))  #.to_table()
)

"""
1) For XCSP competitions, we need to allow AllDifferentList, discard these constraints, or translate them in tables
2) for finding a first solution, the regular model is far more efficient (at least with default heuristics)
  (a few seconds for n=50, 60 or 70)
3) for being compatible with the competition mini-track, we use for the main variant:
   [Sum(x[i, j:j + 3]) >= 1 for i in range(n) for j in range(n - 2)],
   [Sum(x[i, j:j + 3]) <= 2 for i in range(n) for j in range(n - 2)],
4) we can write:
   [Sum(x[i, j:j + 3]) in {1,2} for i in range(n) for j in range(n - 2)], 
"""
