"""
See Minizinc challenge 2018 and 2020 
    "CP and MIP approaches for soccer analysis" by R. Duque, A. Arbelaez and J.-F. Diaz, Journal of industrial and management optimization, 2019
"""

from pycsp3 import *

games, iPoints, positions = data
nGames, nTeams, nPositions = len(games), len(iPoints), len(positions)

pt = [0, 1, 3]

lb_score = min(iPoints[i] + sum(min(pt) for j in range(nGames) if i in games[j]) for i in range(nTeams))
ub_score = max(iPoints[i] + sum(max(pt) for j in range(nGames) if i in games[j]) for i in range(nTeams))

# points[j][0] and points[j][1] are the points for the two teams (indexes 0 and 1) of the jth game
points = VarArray(size=[nGames, 2], dom=pt)

# score[i] is the final score of the ith team
score = VarArray(size=nTeams, dom=range(lb_score, ub_score + 1))

# fp[i] is the final position of the ith team
fp = VarArray(size=nTeams, dom=range(1, nTeams + 1))

# bp[i] is the best possible position of the ith team
bp = VarArray(size=nTeams, dom=range(1, nTeams + 1))

# wp[i] is the worst possible position of the ith team
wp = VarArray(size=nTeams, dom=range(1, nTeams + 1))

satisfy(
    # assigning rights points for each game
    [(points[j, 0], points[j, 1]) in {(0, 3), (1, 1), (3, 0)} for j in range(nGames)],

    # computing final points
    [score[i] - Sum(points[j][0] if i == games[j][0] else points[j][1] for j in range(nGames) if i in games[j]) == iPoints[i] for i in range(nTeams)],

    # computing worst positions (the number of teams with greater total points)
    [wp[i] == Sum(score[j] >= score[i] for j in range(nTeams)) for i in range(nTeams)],

    # computing best positions (from worst positions and the number of teams with equal points)
    [bp[i] == wp[i] - Sum(score[j] == score[i] for j in range(nTeams) if i != j) for i in range(nTeams)],

    # bounding final positions
    (
        [fp[i] >= bp[i] for i in range(nTeams)],
        [fp[i] <= wp[i] for i in range(nTeams)]
    ),

    # ensuring different positions
    AllDifferent(fp),

    # applying rules from specified positions
    (
        [fp[i] == p for i, p in positions],
        [Sum(score[j] > score[i] for j in range(nTeams) if j != i) < p for i, p in positions],
        [Sum(score[j] < score[i] for j in range(nTeams) if j != i) <= nTeams + 1 - p for i, p in positions]
    )
)

""" Comments
1) we can now write iPoints[i] + Sum(...) but it introduces auxiliary variables 
 [iPoints[i] + Sum(points[j][0] if i == games[j][0] else points[j][1] for j in range(numberOfGames) if i in games[j]) == fPoints[i] for i in range(n)]
2) Warning: length(games) in Minizinc returns the number of elements in the array and not the number of rows.
"""
